/*
 * This file is part of SubLib, a library for managing movie subtitles.
 * Copyright (C) 2005-2006 Pedro Castro
 *
 * SubLib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SubLib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

using System;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;

namespace SubLib {
	
/// <summary>Represents the text of a subtitle.</summary>	
public class SubtitleText {
	private ArrayList textLines = new ArrayList();

	/// <summary>Initializes a new instance of the <see cref="SubtitleText" /> class
	/// with the specified text, line break and trim text option.</summary>
	/// <param name="text">The subtitle text.</param>
	/// <param name="lineBreak">The text substring used to split the text in lines.</param>
	/// <param name="toTrim">Whether to trim the text lines.</param>
	public SubtitleText (string text, string lineBreak, bool toTrim) {
		Set(text, lineBreak, toTrim);
	}
	
	/// <summary>Initializes a new instance of the <see cref="SubtitleText" /> class
	/// with the specified text.</summary>
	/// <remarks>Newline (\n) is used as the line break. The text is not trimmed.</remarks>
	/// <param name="text">The subtitle text.</param>
	public SubtitleText (string text) : this(text, "\n", false) {
	}
	
	/// <summary>Initializes a new instance of the <see cref="SubtitleText" /> class.</summary>
	public SubtitleText() {
	}
	
	/// <summary>Whether there is no text.</summary>
	public bool IsEmpty {
		get { return (textLines.Count == 0); }
	}
	
	/// <summary>The number of text lines.</summary>
	public int LineCount {
		get { return textLines.Count; }
	}
	
	/// <summary>Gets the text at the specified line number index.</summary>
	/// <param name="index">The zero-based line number index.</param>
	/// <returns>The specified text line.</returns>
	public string GetLine (int index) {
		if ((index >= 0) && (index < textLines.Count))
			return textLines[index] as string;
		else
			return String.Empty;
	}

	/// <summary>Gets the text lines merged by the specified line break.</summary>
	/// <param name="lineBreak">The line break used to merge the text.</param>
	/// <returns>The subtitle text.</returns>
	public string Get (string lineBreak) {
		string text = String.Empty;
		IEnumerator lines = textLines.GetEnumerator();
		if (lines.MoveNext()){
			text = (string)lines.Current;
			while (lines.MoveNext())
				text += lineBreak + (string)lines.Current;
		}
		return text;
	}
	
	/// <summary>Gets the text lines.</summary>
	/// <remarks>The text lines are merged by the new line (\n) char.</remarks>
	/// <returns>The subtitle text.</returns>
	public string Get () {
		return Get("\n");
	}
	
	/// <summary>Gets the text lines merged by the specified line break and replaces the
	/// empty lines with the specified replacement.</summary>
	/// <param name="replacement">The text to replace empty lines with.</param>
	/// <param name="lineBreak">The line break used to merge the text.</param>
	/// <returns>The subtitle text.</returns>
	public string GetReplaceEmptyLines (string replacement, string lineBreak) {
		string text = String.Empty;
		IEnumerator lines = textLines.GetEnumerator();
		if (lines.MoveNext()){
			string line = (string)lines.Current;
			text = ReplaceLineIfEmpty(line, replacement);
			while (lines.MoveNext()) {
				line = (string)lines.Current;
				text += lineBreak + ReplaceLineIfEmpty(line, replacement);
			}
		}
		return text;
	}
	
	/// <summary>Gets the text lines and replaces empty lines with the specified replacement.</summary>
	/// <param name="replacement">The text to replace empty lines with.</param>
	/// <remarks>The text lines are merged by the new line (\n) char.</remarks>
	/// <returns>The subtitle text.</returns>
	public string GetReplaceEmptyLines (string replacement) {
		return GetReplaceEmptyLines(replacement, "\n");
	}
	
	/// <summary>Sets the subtitle text using the specified line break and trim option.</summary>
	/// <param name="text">The subtitle text.</param>
	/// <param name="lineBreak">The text substring used to split the text in lines.</param>
	/// <param name="toTrim">Whether to trim the text lines.</param>
	public void Set (string text, string lineBreak, bool toTrim) {
		textLines.Clear();
		if (toTrim)
			text = text.Trim();

		string escapedLineBreak = Regex.Escape(lineBreak);
		string spaceDelimiters = (toTrim ? @"\s*" : String.Empty);
		string[] lines = Regex.Split(text, spaceDelimiters + @escapedLineBreak + spaceDelimiters);
		foreach (string line in lines)
			textLines.Add(line);
	}
	
	/// <summary>Sets the subtitle text.</summary>
	/// <remarks>Newline (\n) is used as the line break. The text is not trimmed.</remarks>
	/// <param name="text">The subtitle text.</param>
	public void Set (string text) {
		Set(text, "\n", false);
	}
	
	public override string ToString() {
		string result = String.Empty;
	  	int lineNumber = 1;
  		foreach (string line in textLines){
  			result += "\t" + lineNumber + ". " + line + "\n";
  			lineNumber++;
  		}
  		return result;
	}
	
	/* Private Methods */
	
	private string ReplaceLineIfEmpty (string textLine, string replacement) {
		if (textLine == String.Empty)
			return replacement;
		else
			return textLine;
	}

}

}
